/***********************************************************************/
/*            2009 - Serkan Eryilmaz                                   */
/*            Code libre de droits                                     */
/***********************************************************************/

#include <fstream>
#include <vector>
#include "Klotski.h"

using namespace Klotski;

KlotskiState::KlotskiState( ) :
    m_pParent(NULL),
    m_iG(0)
{

}

/*
 * Deux états sont équivalents si chacunes de ses pièces sont positionnés de la même manière
 */
bool KlotskiState::isEqual( const KlotskiState& k ) const
{
    for ( unsigned short y = 0; y < ROW; y++ )
    {
        for ( unsigned short x = 0; x < COL; x++ )
        {
            if ( m_cBoard[y][x] != k.m_cBoard[y][x] )
            {
                return false;
            }
        }
    }
    return true;
}

/*
 * Deux états sont mirroirs si les pièces sont positionnés de manière à avoir
 * ABCD mirroir de DCBA
 * EFGH            HGFE
 * IJKL            LKJI
 * MNOP            PONM
 */

bool KlotskiState::isEqualOpposite( const KlotskiState& k ) const
{
    for ( unsigned short y = 0; y < ROW; y++ )
    {
        for ( unsigned short x = 0; x < COL; x++ )
        {
            if ( m_cBoard[y][x] != k.m_cBoard[y][COL-1-x] )
            {
                return false;
            }
        }
    }
    return true;
}

/*
 * On remplit notre m_cBoard
 */
void KlotskiState::MakeArray()
{
    for ( std::map<char,Piece>::const_iterator it = m_Board.begin(); it != m_Board.end(); ++it )
    {
        Piece p = it->second;
        switch ( p.type )
        {
        case SQUARE:
        {
            m_cBoard[p.pos.y][p.pos.x] = 'o';
        }
            break;
        case VOID:
            m_cBoard[p.pos.y][p.pos.x] = ' ';
            break;
        case HORIZONTAL:
        {
            m_cBoard[p.pos.y][p.pos.x] = '=';
            m_cBoard[p.pos.y][p.pos.x+1] = '=';
        }
        break;
        case VERTICAL:
        {
            m_cBoard[p.pos.y][p.pos.x] = '|';
            m_cBoard[p.pos.y+1][p.pos.x] = '|';
        }
        break;
        case KLOTSKI:
        {
            m_cBoard[p.pos.y][p.pos.x] = 'X';
            m_cBoard[p.pos.y][p.pos.x+1] = 'X';
            m_cBoard[p.pos.y+1][p.pos.x] = 'X';
            m_cBoard[p.pos.y+1][p.pos.x+1] = 'X';
        }
        break;
        }
    }
}

/*
 * On cherche les pièces A et B (qui sont les pièces de type VIDE) et on compare leur position à la position passée en paramètre.
 * Si on a une équivalence avec A, on renvoi 'A' comme identifiant unique de la pièce vide trouvée
 * Si "  "  "       "        "  B,  "    "   'B'   "      "         "      "  "   "     "     "
 * Si on a pas d'équivalence de positions, alors on renvoit 0
 */
char KlotskiState::isFree( const Position& p ) const
{
    if( m_Board.find('A')->second.pos == p )
        return 'A';

    if( m_Board.find('B')->second.pos == p )
        return 'B';

    return 0;
}

/*
 * Dans cette méthode, on déplace un objet.
 */
void KlotskiState::movePiece( const Piece& piece, const short& xoffset, const short& yoffset, const char& piece_id, const char& void_id )
{
    /*
    Dans cette première partie, on cherche et on met à jour la position de A, B ou les 2, qui sont les pièces de type 'vide'.
    */

    const Type& type = piece.type;
    const short& x   = piece.pos.x;
    const short& y   = piece.pos.y;

    if ( type == SQUARE )
    {
        m_Board[void_id].pos = Position(piece.pos.x,piece.pos.y);
    }
    else if ( type == HORIZONTAL )
    {
        if ( yoffset == 0 )
        {
            m_Board[void_id].pos = Position(x+( xoffset == -1 ? 1 : 0 ) ,y);
        }
        else
        {
            m_Board['A'].pos = Position(x,y);
            m_Board['B'].pos = Position(x+1,y);
        }
    }
    else if ( type == VERTICAL )
    {
        if ( yoffset == 0 )
        {
            m_Board['A'].pos = Position(x,y);
            m_Board['B'].pos = Position(x,y+1);
        }
        else
        {
            m_Board[void_id].pos = Position(x,y+( yoffset == -1 ? 1 : 0 ));
        }
    }
    else if ( type == KLOTSKI )
    {
        if ( yoffset == 0 )
        {
            m_Board['A'].pos = Position(x+(xoffset == -1 ? 1 : 0),y);
            m_Board['B'].pos = Position(x+(xoffset == -1 ? 1 : 0),y+1);
        }
        else
        {
            m_Board['A'].pos = Position(x,y+( yoffset == -1 ? 1 : 0 ));
            m_Board['B'].pos = Position(x+1,y+( yoffset == -1 ? 1 : 0 ));
        }
    }

    // Puis on met à jour la position de la pièce non-vide concernée
    m_Board[piece_id].pos = Position( x+xoffset, y+yoffset );

    // Finalement, on met à jour notre m_cBoard afin d'afficher et d'effectuer les comparaisons correctement
    MakeArray();
}

/*
 * Cette méthode (plutôt lourde, probablement à retravailler) test si un mouvement donné en paramètre (pièce, direction) est possible.
 * Renvoi l'identifiant de la pièce vide concernée si c'est le cas et qu'une seule pièce est concernée
 * Renvoi 1 si 2 pièces sont concernées et que le déplacement est possible
 * Renvoi 0 si le mouvement n'est pas possible
 */
char KlotskiState::movePossible( const Piece& piece, unsigned short direction ) const
{
    char ret;
    switch ( piece.type )
    {
    case SQUARE:
    {
        switch ( direction )
        {
        case LEFT:
        {
            ret = isFree( Position( piece.pos.x-1, piece.pos.y ) );
            if ( piece.pos.x > 0 && ret )
                return ret;
        }
        break;
        case RIGHT:
        {
            ret = isFree( Position( piece.pos.x+1, piece.pos.y ) );
            if ( piece.pos.x < 3 && ret )
                return ret;
        }
        break;
        case UP:
        {
            ret = isFree( Position( piece.pos.x, piece.pos.y-1 ) );
            if ( piece.pos.y > 0 && ret )
                return ret;
        }
        break;
        case DOWN:
        {
            ret = isFree( Position( piece.pos.x, piece.pos.y+1 ) );
            if ( piece.pos.y < 4 && ret )
                return ret;
        }
        break;
        }
    }
    break;
    case HORIZONTAL:
    {
        switch ( direction )
        {
        case LEFT:
        {
            ret = isFree( Position( piece.pos.x-1, piece.pos.y ) );
            if ( piece.pos.x > 0 && ret )
                return ret;
        }
        break;
        case RIGHT:
        {
            ret = isFree( Position( piece.pos.x+2, piece.pos.y ) );
            if ( piece.pos.x+1 < 3 && ret )
                return ret;
        }
        break;
        case UP:
        {
            return piece.pos.y > 0 && isFree( Position( piece.pos.x, piece.pos.y-1 ) ) && isFree( Position( piece.pos.x+1, piece.pos.y-1 ) );
        }
        break;
        case DOWN:
        {
            return piece.pos.y < 4 && isFree( Position( piece.pos.x, piece.pos.y+1 ) ) && isFree( Position( piece.pos.x+1, piece.pos.y+1 ) );
        }
        break;
        }
    }
    break;
    case VERTICAL:
    {
        switch ( direction )
        {
        case LEFT:
            return piece.pos.x > 0 && isFree( Position( piece.pos.x-1, piece.pos.y ) ) && isFree( Position( piece.pos.x-1, piece.pos.y+1 ) );
            break;
        case RIGHT:
            return piece.pos.x < 4 && isFree( Position( piece.pos.x+1, piece.pos.y ) ) && isFree( Position( piece.pos.x+1, piece.pos.y+1 ) );
            break;
        case UP:
        {
            ret = isFree( Position( piece.pos.x, piece.pos.y-1 ) );
            if ( piece.pos.y > 0 && ret )
                return ret;
        }
        break;
        case DOWN:
        {
            ret = isFree( Position( piece.pos.x, piece.pos.y+2 ) );
            if ( piece.pos.y+1 < 4 && ret )
                return ret;
        }
        break;
        }
    }
    break;
    case KLOTSKI:
    {
        switch ( direction )
        {
        case LEFT:
            return piece.pos.x > 0 && isFree( Position( piece.pos.x-1, piece.pos.y ) ) && isFree( Position( piece.pos.x-1, piece.pos.y+1 ) );
            break;
        case RIGHT:
            return piece.pos.x+1 < 3 && isFree( Position( piece.pos.x+2, piece.pos.y ) ) && isFree( Position( piece.pos.x+2, piece.pos.y+1 ) );
            break;
        case UP:
            return piece.pos.y > 0 && isFree( Position( piece.pos.x, piece.pos.y-1 ) ) && isFree( Position( piece.pos.x+1, piece.pos.y-1 ) );
            break;
        case DOWN:
            return piece.pos.y+1 < 4 && isFree( Position( piece.pos.x, piece.pos.y+2 ) ) && isFree( Position( piece.pos.x+1, piece.pos.y+2 ) );
            break;
        }
    }
    break;
    default:
        break;
    }

    return false;
}

KlotskiState& KlotskiState::operator=( const KlotskiState& k )
{
    m_Board = k.m_Board;
    m_pParent = k.m_pParent;
    m_iG = k.m_iG;

    return *this;
}

KlotskiState *KlotskiState::getParent()
{
    return m_pParent;
}

/*
 * Permet d'initialiser un état à partir d'un fichier d'entrée dont le chemin se trouve en paramètre.
 */

bool KlotskiState::Init(const std::string& filename)
{

    std::ifstream InputFile( filename.c_str() );
    std::vector < std::string > vInput;
    if( InputFile )
    {
        std::string curLine;
        unsigned short iLine = 0;

        while( std::getline( InputFile, curLine ) )
        {
            ++iLine;

            if( iLine <= ROW )
            {
                if( curLine.length() == COL )
                {
                    vInput.push_back( curLine );
                }
                else
                {
                    std::cout << "Contenu du fichier invalide.\r\nNombre de caractères sur la ligne " << iLine << " invalide." << std::endl;
                    return false;
                }
            }
            else
            {
                std::cout << "Contenu du fichier invalide.\r\nNombre de lignes invalide." << std::endl;
                return false;
            }
        }
    }
    else
    {
        std::cout << "Impossible d'ouvrir le fichier \"" << filename << "\"" << std::endl;
        return false;
    }

    assert( vInput.size() == ROW );

    for( unsigned short y = 0; y < ROW; y++ )
    {
        for( unsigned short x = 0; x < COL; x++ )
        {
            const char curT = vInput[y][x];
            if( curT == 'A' || curT == 'B' )
            {
                if( m_Board.find( curT ) != m_Board.end() )
                {
                    std::cout << "Case vide " << curT << " retrouvé plus d'une fois." << std::endl;
                    return false;
                }
                else
                {
                    //m_Board.push_back( Piece( Position( x, y ), VOID, curT ) );
                    m_Board[curT] = Piece( Position(x, y), VOID );
                    vInput[y][x] = '_';
                }
            }
            else if( curT == 'C' || curT == 'D' || curT == 'E' || curT == 'F'  )
            {
                if( m_Board.find( curT ) != m_Board.end() )
                {
                    std::cout << "Case carré " << curT << " retrouvé plus d'une fois." << std::endl;
                    return false;
                }
                else
                {
                    //m_Board.push_back( Piece( Position( x, y ), SQUARE, curT ) );
                    m_Board[curT] = Piece( Position(x, y), SQUARE );
                    vInput[y][x] = '_';
                }
            }
            else if( curT == 'G' || curT == 'H' || curT == 'I' || curT == 'J' || curT == 'K' )
            {
                if( m_Board.find( curT ) != m_Board.end() )
                {
                    std::cout << "Case rectangle " << curT << " retrouvé plus d'une fois." << std::endl;
                    return false;
                }
                else
                {
                    if( y < ROW-1 && curT == vInput[y+1][x] )
                    {
                        // vertical
                        //m_Board.push_back( Piece( Position( x,y ), VERTICAL, curT ) );
                        m_Board[curT] = Piece( Position(x, y), VERTICAL );
                        vInput[y][x] = vInput[y+1][x] = '_';

                    }
                    else if( x < COL-1 && curT == vInput[y][x+1] )
                    {
                        // horizontal
                        //m_Board.push_back( Piece( Position( x,y ), HORIZONTAL, curT ) );
                        m_Board[curT] = Piece( Position(x, y), HORIZONTAL );
                        vInput[y][x] = vInput[y][x+1] = '_';
                    }
                    else
                    {
                        // invalid
                        std::cout << "Positionnement de la case rectangle " << curT << " impossible." << std::endl;
                        return false;
                    }
                }
            }
            else if( curT == 'L' )
            {
                if( m_Board.find( curT ) != m_Board.end() )
                {
                    std::cout << "Case rectangle " << curT << " retrouvé plus d'une fois." << std::endl;
                    return false;
                }
                else if( y < ROW-1 && x < COL-1 && vInput[y][x+1] == curT && vInput[y+1][x] == curT && vInput[y+1][x+1] == curT )
                {
                    //m_Board.push_back( Piece( Position( x,y ), KLOTSKI, curT ) );
                    m_Board[curT] = Piece( Position(x, y), KLOTSKI );
                    vInput[y][x] = vInput[y+1][x] = vInput[y][x+1] = vInput[y+1][x+1] = '_';
                }
                else
                {
                    // invalid
                    std::cout << "Positionnement du grand carré incohérent" << std::endl;
                    return false;
                }
            }
            else if( curT != '_' )
            {
                std::cout << "Case inconnue: " << curT << std::endl;
                return false;
            }
        }
    }

    MakeArray();

    return true;
}

int KlotskiState::getCost() const
{
    return m_iG;
}

bool KlotskiState::isGoal() const
{
    return m_Board.find('L')->second.pos == Position(1,3);
}

bool KlotskiState::operator==( const KlotskiState& k ) const
{
    return (isEqual(k) || isEqualOpposite(k));
}

void KlotskiState::Print( ) const
{
    for ( unsigned short i = 0; i < ROW; i++ )
    {
        for ( unsigned short j = 0; j < COL; j++ )
        {
            std::cout << m_cBoard[i][j] << " ";
        }
        std::cout << "\r\n";
    }
    std::cout << std::endl;
}

const std::list< KlotskiState > KlotskiState::getPossibleNextMoves( )
{
    std::list< KlotskiState > ret;
    // we will search for each piece and see what we can do with it
    //std::list< Piece >::iterator piece;
    std::map<char, Piece>::iterator it;

    char free_id = 0;

    for ( it = m_Board.begin(); it != m_Board.end(); ++it )
    {
        char id = it->first;
        if ( id == 'A' || id == 'B' )
            continue;

        Piece& piece = it->second;
        if( piece.pos.x > 0 )
        {
            free_id = movePossible( piece, LEFT );
            if ( free_id )
            {
                KlotskiState m(*this);
                m.m_pParent = this;
                m.m_iG = m_iG+1;
                m.movePiece( piece, -1, 0, id, free_id );
                if ( getParent() == NULL || !( m == *getParent() )  )
                    ret.push_back(m);
            }
        }
        if( piece.pos.x < COL-1 )
        {
            free_id = movePossible( piece, RIGHT );
            if ( free_id )
            {
                KlotskiState m(*this);
                m.m_pParent = this;
                m.m_iG = m_iG+1;
                m.movePiece( piece, 1, 0, id, free_id );
                if ( getParent() == NULL || !( m == *getParent() )  )
                    ret.push_back(m);
            }
        }
        if( piece.pos.y > 0 )
        {
            free_id = movePossible( piece, UP );
            if ( free_id )
            {
                KlotskiState m(*this);
                m.m_pParent = this;
                m.m_iG = m_iG+1;
                m.movePiece( piece, 0, -1, id, free_id );
                if ( getParent() == NULL || !( m == *getParent() )  )
                    ret.push_back(m);
            }
        }
        if( piece.pos.y < ROW-1 )
        {
            free_id = movePossible( piece, DOWN );
            if ( free_id )
            {
                KlotskiState m(*this);
                m.m_pParent = this;
                m.m_iG = m_iG+1;
                m.movePiece( piece, 0, 1, id, free_id );
                if ( getParent() == NULL || !( m == *getParent() )  )
                    ret.push_back(m);
            }
        }
    }
    return ret;
}

