/***********************************************************************/
/*            2009 - Serkan Eryilmaz                                   */
/*            Code libre de droits                                     */
/***********************************************************************/

#ifndef SOLVER_H_INCLUDED
#define SOLVER_H_INCLUDED

#include <iostream>
#include <deque>
#include <list>
#include <boost/shared_ptr.hpp>
#include <boost/progress.hpp>

namespace Solver
{
    /*
    Solver::State
    Interface générique d'un état duquel on va pouvoir chercher une solution.
    */
    template< class T>
    class State
    {
    public:

        typedef std::list< T > StateList;

        /*
        Cette méthode permet au moteur de recherche de connaître le coût,
        à partir de l'état initiale, du chemin parcouru jusqu'à l'état courant.
        */
        virtual int getCost() const = 0;

        /*
        Cette méthode permet au moteur de recherche de savoir si l'état courant
        une solution valable.
        */
        virtual bool isGoal() const = 0;

        /*
        Le moteur de recherche doit pouvoir comparer plusieurs états afin de ne
        jamais se retrouver 2 fois sur le même état, donc risquer de tourner en
        boucle au pire, et de chercher très longtemps au mieux.
        */
        virtual bool operator==( const T& ) const = 0;

        /*
        Le moteur de recherche a besoin de faire ses optimisations internes,
        sans pour autant désallouer et réallouer. Il doit simplement pouvoir
        affecter à un état l'état d'un autre état.
        */
        virtual T& operator=( const T& ) = 0;

        /*
        Cette méthode doit permettre l'affichage de l'état courant sous forme
        texte. Il peut ne pas être implémenté (laissé vide), mais il doit exister.
        */
        virtual void Print( ) const = 0;

        /*
        Le moteur doit connaître les états successeurs de l'état courant, afin
        de pouvoir continuer à chercher. Cette méthode permet de retourner les
        successeurs de l'état courant.
        */
        virtual const StateList getPossibleNextMoves( ) = 0;

        /*
        Lorsqu'on a trouvé un chemin, il faut pouvoir remonter jusqu'à l'état
        initial. Cette méthode permettra via une fonction récursive de remonter
        et d'afficher le chemin trouvé.
        */
        virtual T* getParent() = 0;
    };

    /*
    Les différents algorithmes implémentés sont énumérées ici.
    */
    enum Algo
    {
        DEPTH,      // Algorithme de recherche et de sélection par profondeur
        BREADTH,    // Algorithme de recherche et de sélection par largeur
        HEURISTIC   // Algorithme de recherche et de sélection par coût
    };

    /*
    Ceci est le pattern définissant l'algorithme de sélection d'état parmis les
    états successeurs proposés.

    De base, la sélection n'est pas implémentée et doit être implémentée par une
    classe dérivée de celle-ci.

    La méthode d'expansion d'état est cependant implémentée car elle peut-être
    commune à plusieurs algorithmes. Cependant, rien n'empêche de la ré-implémenter
    dans la classe dérivée.
    */
    template< class T>
    class AlgorithmAbstract
    {
        /*
        Cette variable contiendra le coût maximale
        à ne pas dépasser lors de la recherche
        */
        int m_iMaxCost;

        typedef boost::shared_ptr< State< T > > StatePtr;
    public:

        AlgorithmAbstract( int maxcost ) :
                m_iMaxCost( maxcost ) {}

        /*
        Non implémentée de base.
        Cette méthode doit être implémentée par les classes dérivées.

        Elle prend en paramètres la queue d'états ouvertes, c'est à dire celles
        qu'on a jusque là pas encore visitées, et une liste d'état fermées, c'est à dire
        celles qu'on a déjà visitées et probablement extrait.

        Cette méthode doit extraire un état, selon le critère de son choix, du premier
        paramètre, doit l'insérer dans la seconde liste d'états et renvoyer l'état
        sélectionné.
        */
        virtual StatePtr select( std::deque< StatePtr >&, std::list< StatePtr >& ) = 0;

        /*
        Cette méthode extrait simplement les états successeurs d'un état n (premier paramètre),
        vérifie que chacun de ces états n'ait pas déjà été visitées et qu'il ne soit pas déjà prévu
        à être visité, et si c'est le cas, l'insère dans la liste des états qu'on pourrait visiter.

        Ceci en prenant en compte le coût des états successeurs, qui ne doit pas dépasser le coût
        maximal fixé à l'avance (ou pas), par l'utilisateur de la classe.
        */
        virtual void expand( StatePtr& n, std::deque< StatePtr >& open, std::list< StatePtr >& closed )
        {
            /*
            Nous récupérons ici les états successeurs de l'état n.
            N'oublions pas que getPossibleNextMoves() doit être implémenté par la classe
            définissant un état, qui est un dérivé de l'interface State.
            */
            std::list< T > Successors = n->getPossibleNextMoves();
            typename std::list< T >::const_iterator it;

            /*
            Début de l'itération des successeurs de n
            */
            for ( it = Successors.begin(); it != Successors.end(); ++it )
            {
                /*
                Pour chaque successeur m de n
                */
                boost::shared_ptr< T > m( new T( *it ) );

                // Si un coût a été fixé et que le coût de l'état m ne dépasse pas ce coût
                if ( m_iMaxCost == -1 || m->getCost() < m_iMaxCost )
                {
                    // notre variable déterminant si on a trouvé ou pas l'état dans l'une de nos
                    // conteneurs d'états.
                    bool found = false;

                    typename std::deque< StatePtr >::iterator jt;

                    // Nous lançons la recherche sur notre conteneurs d'états "ouverts".
                    for ( jt = open.begin(); jt != open.end(); ++jt )
                    {
                        if ( *(*jt) == *m )
                        {
                            // Si nous avons trouvé un état équivalent
                            found = true;

                            // Et que l'état m nous a coûte moins cher que l'état qu'on souhaitait visiter
                            if ( (*jt)->getCost() > m->getCost() )
                            {
                                // alors nous diront qu'il y a un chemin jusqu'à l'état m
                                // qui coûte moins cher que celui que nous avions déjà visité.
                                *jt = m;

                                // Ne pas oublier que l'opérateur d'assignation doit être implémenté par
                                // la classe fille dérivant de l'interface State.
                                // Et que cette implémentation doit prendre en compte la gestion du coût.
                            }

                            // Etant donné qu'on l'a déjà trouvé, il est inutile de continuer à chercher.
                            break;
                        }
                    }

                    // Il est évident que si on l'a trouvé dans le conteneurs d'états "ouverts",
                    // il ne devrait pas être dans le conteneurs d'états "fermés".
                    // Afin de ne pas effectuer de recherches inutiles, nous éliminons la recherche
                    // si on l'a déjà trouvé.
                    if( !found )
                    {
                        // Sinon
                        typename std::list< StatePtr >::iterator yt;
                        for ( yt = closed.begin(); yt != closed.end(); ++yt )
                        {
                            if ( *(*yt) == *m )
                            {
                                // On a retrouvé l'état m dans la liste d'états fermés.
                                found = true;

                                // N'oublions pas que ces états peuvent faire partie de la
                                // solution optimale.
                                if ( (*yt)->getCost() > m->getCost() )
                                {
                                    // Si le coût du chemin que nous avons visité précédemment
                                    // est supérieur au coût de l'état courant m,
                                    // alors on peut supposer que l'état m peut à nouveau être
                                    // intéressant.

                                    // Ceci est afin d'éviter qu'un chemin pouvant être le chemin
                                    // optimal ne soit jamais visité car son coût précédent était
                                    // trop élevé.

                                    // Nous allons donc enlever l'état "cher" de notre conteneur
                                    // d'états fermés.
                                    closed.erase(yt);

                                    // Et nous allons supposer qu'on a pas trouvé l'état m dans les
                                    // conteneurs.
                                    found = false;
                                }

                                break;
                            }
                        }
                    }

                    // Si nous n'avons pas retrouvé l'état, ou s'il nécessite d'être mis à jour,
                    if ( !found )
                    {
                        // Alors nous l'insérons dans la liste d'états à visiter.
                        open.push_back( m );
                    }
                }
            }
        }
    };

    /*
    Définition de l'algorithme de sélection par profondeur-en-premier.
    */
    template<class T >
    class DepthFirstAlgo : public AlgorithmAbstract< T >
    {
    public:

        typedef boost::shared_ptr< State< T > > StatePtr;

        DepthFirstAlgo( const int limit ) :
                AlgorithmAbstract< T >( limit )
        {}

        // Sachant que les états successeurs sont insérés à la fin du conteneur,
        // Il suffit de récupérer un état qui se trouve tout derrière.
        // Sa profondeur sera supérieure à l'état précédent.
        StatePtr select( std::deque< StatePtr >& open, std::list< StatePtr >& closed)
        {
            StatePtr n = open.back();
            closed.push_back( n );
            open.pop_back();

            return n;
        }
    };

    /*
    Définition de l'algorithme de sélection par largeur-en-premier
    */
    template<class T >
    class BreadthFirstAlgo : public AlgorithmAbstract< T >
    {
    public:
        typedef boost::shared_ptr< State< T > > StatePtr;

        BreadthFirstAlgo(const int limit) :
                AlgorithmAbstract< T >( limit )
        {}

        /*
        De manière similaire à l'algorithme de recherche par profondeur,
        Cet algorithme va plutôt prendre les états se trouvant en début du
        conteneur car ceux-ci sont ceux dont la profondeur est la moins élevée.
        */
        StatePtr select( std::deque< StatePtr >& open, std::list< StatePtr >& closed)
        {
            StatePtr n = open.front();
            closed.push_back( n );
            open.erase(open.begin());

            return n;
        }
    };

    /*
    Définition de l'algorithme de sélection par coût.
    */
    template<class T >
    class HeuristicAlgo : public AlgorithmAbstract< T >
    {
    public:

        typedef boost::shared_ptr< State< T > > StatePtr;

        HeuristicAlgo(const int limit) :
                AlgorithmAbstract< T >( limit )
        {}

        StatePtr select( std::deque< StatePtr >& open, std::list< StatePtr >& closed)
        {
            /*
            Nous allons, ici, pour sélectionner un état, effectuer un prédicat.
            Tout simplement en admettant que l'état sélectionné courant est
            l'état qui coûte le moins cher.

            Nous allons d'abord sélectionner le premier état de la queue d'états.
            Etant donné que nous sélectionnons, nous devons l'enlever de la queue
            d'états ouverts et l'insérer dans la liste d'états fermés.

            Nous pouvons de manière sûre sélectionner le premier état. Sachant que
            si on nous demande de sélectionner un état, on suppose que la queue d'états
            duquel on peut sélectionner n'est pas vide.

            Par précaution, nous allons effectuer une assertion.
            */

            assert( open.size() > 0 );

            StatePtr n = open.front();
            closed.push_back( n );
            open.erase(open.begin());

            typename std::deque< StatePtr >::iterator it;

            /*
            Il est inutile d'itérer à nouveau sur le premier élément, nous
            l'avons déjà sélectionné.
            */
            for ( it = open.begin()+1; it != open.end(); ++it )
            {
                if ( (*it)->getCost() < n->getCost() )
                {
                    /*
                    Si nous avons trouvé un état qui coûterai moins cher que l'état sélectionné courant,
                    alors nous mettons à jour l'état courant afin de garder l'invariant correct.
                    */

                    // Nous remettons l'état précédemment sélectionné dans la queue d'états ouverts,
                    // étant donné que nous n'allons pas la visiter pour l'instant
                    open.push_back( n );

                    // Nous enlevons l'état de la liste d'états fermés, étant donné que nous ne la
                    // sélectionnons plus
                    closed.pop_back();

                    // Et nous resélectionnons l'état.
                    n = (*it);
                    open.erase( it );
                    closed.push_back( n );
                }
            }

            // Retourne l'état le moins cher.
            return n;
        }
    };

    /*
    Notre classe "Moteur".
    Ceci est le coeur de l'algorithme de recherche.
    Elle est générique et peut-être appliquée à plus ou moins n'importe quel
    recherche. Ceci est une implémentation acceptant comme paramètre template
    un gestionnaire d'état, donc un dérivé de l'interface State.
    */
    template< class T >
    class Solver
    {
        typedef boost::shared_ptr< State< T > > StatePtr;

        // Notre conteneur qui va contenir les états intéressant à visiter
        // J'ai utilisé un std::deque car il permet de sélectionner les états
        // en début et en fin de queue.
        std::deque< StatePtr > open;

        // La liste qui contient les états déjà visités.
        std::list< StatePtr > closed;

        // L'état de notre algorithme de sélection et d'expansion.
        AlgorithmAbstract<T> *m_Algo;

        // Cette variable va contenir le coût du chemin trouvé
        // à partir de l'état initial jusqu'à l'état final/
        int m_iCost;

        // Cette variable va contenir le nombre d'itérations effectuées
        // afin d'obtenir la solution
        int m_iIterations;

        // Cette méthode récursive permet de retracer le chemin trouvé
        void PrintPath( T* node )
        {
            if ( node->getParent() != NULL )
            {
                PrintPath( node->getParent() );
            }
            std::cout << "Etape: " << node->getCost() << "\r\n";
            node->Print();
        }

        // La méthode de recherche principale.
        void Run( )
        {
            // Afin de pouvoir effectuer une recherche,
            // nous devons un et un seul état à visiter.
            // Cet état est l'état initial
            assert ( open.size() == 1 );

            // Nous le sélectionnons afin de pouvoir l'afficher uniquement
            // Nous ne le visitons pas encore, donc nous ne l'enlevons pas
            // Et nous ne l'insérons pas dans la liste d'états "fermées".
            StatePtr s = open.front();
            std::cout << "Etat initial: \r\n";
            s->Print();

            // progress display affiche simplement le coût actuel de la recherche
            boost::progress_display pd(100, std::cout, "Recherche...\r\n");

            // Tant que nous avons des états à visiter et que nous n'avons pas encore trouvé
            // de chemin, nous chercons.
            while ( open.size() > 0 && m_iCost == -1 )
            {
                m_iIterations++;

                /*
                Nous sélectionnons un état avec l'algorithme choisi
                */
                StatePtr n;
                n = m_Algo->select( open, closed );

                // Nous nous assurons qu'un état a bien été sélectionné.
                assert( n != 0 );

                static int lastCost = n->getCost();

                /*
                Ceci est uniquement pour l'affichage de la progression
                */
                if ( lastCost != n->getCost() )
                {
                    pd += n->getCost() - lastCost;
                    lastCost = n->getCost();
                }

                /*
                Si l'état sélectionné est un état final
                */
                if ( n->isGoal() )
                {
                    // Alors nous affichons le chemin jusqu'à cet état final.
                    PrintPath( n->getParent() );
                    std::cout << "Etat final :\r\n";
                    n->Print();

                    m_iCost = n->getCost();
                    std::cout << "Nombre d'étapes nécessaires: " << m_iCost << std::endl;

                    // Et nous arrêtons.
                    return;
                }

                // Sinon, nous continuons à chercher d'autres états intéressants.
                m_Algo->expand( n, open, closed );
            }
        }

    public:

        Solver() :
                m_Algo(NULL)
        {
        }

        ~Solver()
        {
            delete m_Algo;
        }

        int Solve( T& start_state, const int flag, const int limit = -1 )
        {
            /*
            Nous insérons dans notre conteneurs ouverts notre état initial,
            afin de pouvoir lancer une recherche.
            */
            StatePtr s(new T(start_state));
            open.push_back( s );

            /*
            Nous n'avons pas encore trouvé de solution, donc nous n'avons pas de
            coût associé.
            */
            m_iCost = -1;
            m_iIterations = 0;

            /*
            Nous choisissons notre algorithme de sélection et d'expansion
            */
            if ( m_Algo != NULL )
            {
                delete m_Algo;
                m_Algo = NULL;
            }

            if ( flag == DEPTH )
            {
                m_Algo = new DepthFirstAlgo<T>(limit);
            }
            else if ( flag == BREADTH )
            {
                m_Algo = new BreadthFirstAlgo<T>(limit);
            }
            else
            {
                // heuristic
                m_Algo = new HeuristicAlgo<T>(limit);
            }

            // Nous lançons la recherche
            Run();

            // La recherche est terminée, nous pouvons vider nos conteneurs.
            open.clear();
            closed.clear();

            // Et nous renvoyons le coût du chemin, si trouvé, sinon, le retour
            // sera de -1
            return m_iCost;
        }

        int getNumIterations()
        {
            return m_iIterations;
        }
    };
}

#endif // SOLVER_H_INCLUDED
